
import Debug from 'debug'
import cloneDeep from 'lodash/cloneDeep'
import immediate from 'p-immediate'

/**
 * Debug.
 *
 * @type {Function}
 */

const debug = Debug('nexus:store-memory')

/**
 * The default, memory-only store adapter.
 *
 * @type {MemoryStore}
 */

class MemoryStore {

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  constructor(config = {}) {
    this.snapshots = config.snapshots || {}
    this.operations = config.operations || {}
  }

  /**
   * Save an `operation` and `snapshot` on a document by `id`.
   *
   * @param {String} id
   * @param {Object} operation
   * @param {Object} snapshot
   * @param {Object} options
   * @return {Boolean}
   */

  async commit(id, operation, snapshot, options) {
    debug('commit', { id, operation, snapshot, options })

    await immediate()

    const ops = this.operations[id] || (this.operations[id] = [])
    const version = (ops && ops.length) || 0
    if (snapshot.version !== version + 1) return false

    ops[operation.version] = cloneDeep(operation)
    this.snapshots[id] = cloneDeep(snapshot)
    return true
  }

  /**
   * Find the latest snapshot of a document by `id`.
   *
   * @param {String} id
   * @param {Object} options
   * @return {Object}
   */

  async findSnapshot(id, options) {
    debug('findSnapshot', { id, options })

    await immediate()

    const snapshot = this.snapshots[id]
    if (!snapshot) return

    snapshot.data = cloneDeep(snapshot.data)
    return snapshot
  }

  /**
   * List the operations for a document between versions `from` and `to`. The
   * operations should include `from` and include everything up to, but NOT
   * including `to`.
   *
   * @param {String} id
   * @param {Number} from
   * @param {Number} to
   * @param {Object} options
   * @return {Array}
   */

  async listOperations(id, from, to, options) {
    debug('listOperations', { id, from, to, options })

    await immediate()

    const ops = this.operations[id]
    if (!ops) return []

    const slice = cloneDeep(ops.slice(from, to))
    return slice
  }

}

/**
 * Export.
 *
 * @type {MemoryStore}
 */

export default MemoryStore
