
import Debug from 'debug'
import Emitter from 'events'

import Document from './document'
import { checkMessage, checkTransport, checkTypes } from '../utils/checkers'

/**
 * Debug.
 *
 * @type {Function}
 */

const debug = Debug('nexus:client')

/**
 * Client.
 *
 * @type {Client}
 */

class Client extends Emitter {

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  constructor(config = {}) {
    super()

    const { transport, types } = config
    checkTransport(transport)
    checkTypes(types)

    this.transport = transport
    this.types = types
    this.documents = {}
    this.id = null
    this.sequence = 1
    this.connected = false

    transport.on('open', this.onTransportOpen)
    transport.on('close', this.onTransportClose)
    transport.on('readable', this.onTransportReadable)
    transport.on('error', this.onTransportError)

    if (!transport.closed) {
      this.onTransportOpen()
    }
  }

  /**
   * Open a new document by `id`.
   *
   * @param {String} id
   * @return {Document}
   */

  open(id) {
    debug('open', id)

    let document = this.documents[id]
    if (document) return document

    const { types } = this
    document = new Document({ id, types, client: this })
    this.documents[id] = document

    document.once('close', () => {
      delete this.documents[id]
    })

    return document
  }

  /**
   * Wait for all of the client's document's inflight pending to complete.
   */

  async settle() {
    const promises = Object.values(this.documents).map(d => d.settle())
    await Promise.all(promises)
  }

  /**
   * Send a `message` to the server.
   *
   * @param {Object} message
   */

  send(message) {
    if (!this.connected && message.action !== 'connect') {
      return this.once('connect', () => {
        this.send(message)
      })
    }

    debug('send', message)

    checkMessage(message)
    this.transport.write(message)
  }

  /**
   * On transport open.
   */

  onTransportOpen = () => {
    this.send({ action: 'connect' })
  }

  /**
   * On transport close.
   *
   * @param {String} reason
   */

  onTransportClose = (reason) => {
    this.sequence = 1
    this.id = null
    this.connected = false
    this.emit('disconnect', reason)
  }

  /**
   * On transport readable.
   */

  onTransportReadable = async () => {
    const message = this.transport.read()
    checkMessage(message)

    debug('onMessage', { message })

    if (message.action === 'connect') {
      this.id = message.id
      this.connected = true
      this.emit('connect')
      return
    }

    const doc = this.documents[message.document]
    if (!doc) return

    try {
      await doc.onMessage(message)
    } catch (e) {
      this.emit('error', e)
      return
    }
  }

  /**
   * On transport error.
   *
   * @param {Error} error
   */

  onTransportError = (error) => {
    this.emit('error', error)
  }

}

/**
 * Export.
 *
 * @type {Client}
 */

export default Client
