
import Debug from 'debug'
import Emitter from 'events'
import defer from 'p-defer'

import NexusError from '../utils/nexus-error'

import {
  checkMessage,
  checkSnapshot,
  checkType,
  checkTypes,
} from '../utils/checkers'

/**
 * Debug.
 *
 * @type {Function}
 */

const debug = Debug('nexus:document')

/**
 * Document.
 *
 * @type {Document}
 */

class Document extends Emitter {

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  constructor(config = {}) {
    super()

    const { id, client, types } = config
    checkTypes(types)

    this.id = id
    this.version = null
    this.type = null
    this.value = undefined

    this.types = types
    this.deferreds = {}
    this.queue = []
    this.inflight = null
    this.closed = false
    this.subscribed = false

    this.client = client
    this.client.on('disconnect', this.onClientDisconnect)
  }

  /**
   * Close the document.
   */

  async close() {
    debug('close', this.id)

    await this.settle()
    await this.unsubscribe()

    this.client.removeListener('disconnect', this.onClientDisconnect)
    this.closed = true
    this.emit('close')
  }

  /**
   * Wait for the document's pending messages to complete.
   */

  async settle() {
    const { fetch, subscribe, unsubscribe, queue } = this.deferreds

    await Promise.all([
      fetch && fetch.promise,
      subscribe && subscribe.promise,
      unsubscribe && unsubscribe.promise,
      queue && queue.promise,
    ])
  }

  /**
   * Fetch the document from the server. Unlike subscribe, this is only a
   * one-time fetch that won't listen for subsequent operation messages.
   *
   * @param {Object} options
   * @return {Promise}
   */

  fetch(options = {}) {
    debug('fetch', this.id)

    const { force = false } = options

    if (!this.deferreds.fetch || force) {
      this.send({
        action: 'fetch',
        document: this.id,
        version: this.version,
      })
    }

    const deferred = this.createDeferred('fetch')
    return deferred.promise
  }

  /**
   * Subscribe to a document, fetching it from the server and listening for any
   * subsequent operations that are submitted by other connected clients.
   *
   * @param {Object} options
   * @return {Promise}
   */

  subscribe(options = {}) {
    debug('subscribe', this.id)

    const { force = false } = options

    if (!this.deferreds.subscribe || force) {
      this.send({
        action: 'subscribe',
        document: this.id,
        version: this.version,
      })
    }

    const deferred = this.createDeferred('subscribe')

    // Update the `subscribed` flag when the deferred resolves, so that we can
    // know whether to try re-subscribing on reconnect.
    deferred.promise.then((value) => {
      this.subscribed = true
      return value
    })

    return deferred.promise
  }

  /**
   * Unsubscribe from the document, which stops listening for new operations.
   *
   * @param {Object} options
   * @return {Promise}
   */

  unsubscribe(options = {}) {
    debug('unsubscribe', this.id)

    const { force = false } = options

    // Update the `subscribed` flag optimistically, in case the message errors.
    this.subscribed = false

    if (!this.deferreds.unsubscribe || force) {
      this.send({
        action: 'unsubscribe',
        document: this.id,
      })
    }

    const deferred = this.createDeferred('unsubscribe')
    return deferred.promise
  }

  /**
   * Send a `message` to the server.
   *
   * @param {Object} message
   */

  send(message) {
    if (this.closed) {
      throw new Error(`Tried to send a message for document "${this.id}", but it was already closed.`)
    }

    debug('send', this.id, { message })
    this.client.send(message)
  }

  /**
   * Submit an `operation` to the server by adding it to the operation queue.
   *
   * @param {Object} operation
   */

  submit(operation) {
    const { type, version, id, queue, value } = this

    if (type == null || version == null) {
      throw new Error('You must call `document.fetch()` or `document.subscribe()` before you can submit operations to a document.')
    }

    debug('submit', id, { operation })

    if (type.normalize) {
      operation = type.normalize(operation)
    }

    const last = queue[queue.length - 1]
    const message = {
      action: 'submit',
      document: id,
      data: operation,
    }

    // If the OT type defines a `compose` method, it must be able to compose any
    // operation into another, so we can compose into the last queued instead.
    if (type.compose && last) {
      last.data = type.compose(last.data, message.data)
    }

    // Otherwise, just add the message to the queue.
    else {
      queue.push(message)
      this.createDeferred('queue')
    }

    // Regardless, apply the non-composed message, since even if we're composing
    // the last was already applied individually.
    this.value = type.apply(value, message.data)

    // Delay flushing to allow composing multiple queued operations.
    process.nextTick(() => this.flush())
  }

  /**
   * Flush any queued operations to the server.
   *
   * @param {Object} options
   */

  flush(options = {}) {
    const { force = false } = options

    if (!this.client.connected) return
    if (this.deferreds.operation && !force) return

    const message = this.inflight || this.queue.shift()

    if (!message) {
      this.resolveDeferred('queue')
      return
    }

    debug('flush', this.id, { message })

    // Since the message could have been an inflight one that got rejected, we
    // need to set the version right before flushing, not when queueing.
    message.version = this.version

    // Do not overwrite an existing `id`, in case this is a retry of an
    // existing operations or the client has reconnected and gotten a new ID.
    if (message.id == null) {
      message.id = `${this.client.id}:${this.client.sequence++}`
    }

    this.inflight = message
    this.createDeferred('submit')
    this.send(message)
  }

  /**
   * Update the document's version and value with a `snapshot`.
   *
   * @param {Object} snapshot
   */

  async update(snapshot) {
    const { data, version } = snapshot

    checkSnapshot(snapshot)
    checkType(snapshot.type, this.types)

    const type = this.types.find(t => t.name === snapshot.type)

    // If we've already ingested the initial snapshot, we can't do it again. We
    // need to fetch operations to transform and apply to the existing snapshot.
    // Although if the version is old, we can just ignore it.
    if (this.type) {
      if (version > this.version) await this.fetch()
      return
    }

    this.version = version
    this.type = type
    this.value = type.deserialize ? type.deserialize(data) : data
  }

  /**
   * Create a deferred object with `action`.
   *
   * @param {String} action
   */

  createDeferred(action) {
    let deferred = this.deferreds[action]
    if (deferred) return deferred

    debug('createDeferred', { action })

    deferred = this.deferreds[action] = defer()
    return deferred
  }

  /**
   * Resolve a deferred object by `action` with `value`.
   *
   * @param {String} action
   * @param {Mixed} value
   */

  resolveDeferred(action, value) {
    const deferred = this.deferreds[action]
    if (!deferred) return

    debug('resolveDeferred', { action, value })

    deferred.resolve(value)
    delete this.deferreds[action]
  }

  /**
   * Reject a deferred object by `action` with `err`.
   *
   * @param {String} action
   * @param {Mixed} err
   */

  rejectDeferred(action, err) {
    const deferred = this.deferreds[action]
    if (!deferred) return

    debug('rejectDeferred', { action, err })

    deferred.reject(err)
    delete this.deferreds[action]
  }

  /**
   * On message.
   *
   * @param {Object} message
   */

  async onMessage(message) {
    debug('onMessage', this.id, { message })

    checkMessage(message)

    let err = null
    const { action, error } = message

    if (error) {
      err = new NexusError('server_error', { error })
      err.code = error.code
      err.data = message
    }

    switch (action) {
      case 'broadcast': return await this.onBroadcast(err, message)
      case 'fetch': return await this.onFetch(err, message)
      case 'submit': return await this.onSubmit(err, message)
      case 'subscribe': return await this.onSubscribe(err, message)
      case 'unsubscribe': return await this.onUnsubscribe(err, message)
    }
  }

  /**
   * On "broadcast" message.
   *
   * @param {Object} err
   * @param {Object} message
   */

  async onBroadcast(err, message) {
    const { inflight, type, value, queue } = this
    const { version, data } = message

    // If the document isn't yet fetched, the version is unset, so fetch it.
    if (this.version == null) {
      await this.fetch()
      return
    }

    // If the version is in the future, re-fetch the document. This can happen
    // when an operation message is missed for some reason from the server.
    if (version > this.version) {
      await this.fetch()
      return
    }

    // If the version is old, ignore it because it's a duplicate. This can
    // happen when an operation from another client is committed right before
    // one we've sent, resulting in it being broadcast but also returned from
    // the submit attempt.
    if (version < this.version) {
      return
    }

    // Otherwise, this is a new operation from another connected client. So
    // transform local operations with it, increment and apply it.
    const locals = queue.slice()
    if (inflight) locals.unshift(inflight)

    for (const local of locals) {
      const l = type.transform(local.data, message.data, 'left')
      const s = type.transform(message.data, local.data, 'right')
      local.data = l
      message.data = s
    }

    this.value = type.apply(value, message.data)
    this.version++
    this.emit('operation', data, { source: 'remote' })
  }

  /**
   * On "fetch" message.
   *
   * @param {Object} err
   * @param {Object} message
   */

  async onFetch(err, message) {
    if (err) return this.rejectDeferred('fetch', err)

    const { snapshot } = message
    if (snapshot) await this.update(snapshot)
    this.resolveDeferred('fetch', this.value)
    if (snapshot) this.emit('fetch', this.value)
  }

  /**
   * On "submit" message.
   *
   * @param {Object} err
   * @param {Object} message
   */

  async onSubmit(err, message) {
    const { inflight } = this
    const { id, version, data } = message

    // If there was an error, the server rejected it. To get back into a normal
    // state, we need to re-fetch the document.
    if (err) {
      this.type = null
      this.version = null
      this.inflight = null
      this.queue = []
      await this.fetch()

      this.rejectDeferred('submit', err)
      return
    }

    // If there isn't an inflight operation, something went wrong.
    if (!inflight) {
      const e = new Error(`Unexpectedly received acknowledgement for an operation when none was in flight.`)
      this.rejectDeferred('submit', e)
      return
    }

    // If the ID's don't match up, something went wrong.
    if (id !== inflight.id) {
      const e = new Error(`Unexpectedly received acknowledgement for an operation with an ID that did not match the current in flight operation's ID. Expected "${inflight.id}", but received "${id}".`)
      this.rejectDeferred('submit', e)
      return
    }

    // If the version is newer than the current version, something went wrong.
    if (version > this.version) {
      const e = new Error(`Unexpectedly received acknowledgement for an operation with a version greater than the document's current version. Expected: "${this.version}", but received "${version}".`)
      this.rejectDeferred('submit', e)
      return
    }

    // If the version received was less than expected, it's old. This can
    // happen when another operation was committed before ours was and we've
    // since already received the broadcast of that other operation.
    if (version < this.version) {
      await this.fetch()
      this.resolveDeferred('submit')
      return
    }

    this.version++
    this.inflight = null
    this.resolveDeferred('submit')
    this.flush()
    this.emit('operation', data, { source: 'local' })
  }

  /**
   * On "subscribe" message.
   *
   * @param {Object} err
   * @param {Object} message
   */

  async onSubscribe(err, message) {
    if (err) return this.rejectDeferred('subscribe', err)

    const { snapshot } = message
    if (snapshot) await this.update(snapshot)
    this.resolveDeferred('subscribe', this.value)
    if (snapshot) this.emit('subscribe', this.value)
  }

  /**
   * On "unsubscribe" message.
   *
   * @param {Object} err
   */

  async onUnsubscribe(err) {
    if (err) return this.rejectDeferred('unsubscribe', err)

    this.resolveDeferred('unsubscribe')
    this.emit('unsubscribe')
  }

  /**
   * On client disconnect.
   */

  onClientDisconnect = () => {
    const { inflight, deferreds, client, queue, subscribed } = this
    const hasPendingSubscribe = !!deferreds.subscribe
    const hasPendingFetch = !!deferreds.fetch

    // If there's a pending inflight operation, return it to the queue.
    if (inflight) {
      queue.unshift(inflight)
      this.inflight = null
    }

    this.resolveDeferred('unsubscribe')

    // When the client reconnects, force re-send the messages, so that any
    // original deferreds can be resolved/rejected again.
    client.once('connect', () => {
      this.flush({ force: true })
      if (hasPendingSubscribe || subscribed) this.subscribe({ force: true })
      if (hasPendingFetch) this.fetch({ force: true })
    })
  }

}

/**
 * Export.
 *
 * @type {Document}
 */

export default Document
