
import Debug from 'debug'
import Emitter from 'events'
import hat from 'hat'
import omit from 'lodash/omit'

import { checkMessage } from '../utils/checkers'

/**
 * Debug.
 *
 * @type {Function}
 */

const debug = Debug('nexus:connection')

/**
 * Connection.
 *
 * @type {Connection}
 */

class Connection extends Emitter {

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  constructor(config = {}) {
    super()

    const { server, transport, context } = config
    this.server = server
    this.transport = transport
    this.context = context
    this.id = hat()
    this.subscriptions = {}

    transport.on('readable', () => {
      const data = this.transport.read()
      if (data === null) return this.close()
      this.onMessage(data)
    })
  }

  /**
   * Close the connection.
   */

  close() {
    const { subscriptions } = this

    for (const key in subscriptions) {
      const sub = subscriptions[key]
      sub.push(null)
    }

    this.emit('close')
  }

  /**
   * Send a `message` to the client.
   *
   * @param {Object} message
   */

  send(message) {
    debug('send', this.id, { message })
    this.transport.write(message)
  }

  /**
   * Send a connect message to the client for a `document`.
   *
   * @param {String} document
   */

  sendConnect(document) {
    this.send({
      action: 'connect',
      id: this.id,
    })
  }

  /**
   * Send a `snapshot` message to the client for a `document` and `action`.
   *
   * @param {String} document
   * @param {String} action
   * @param {Object} snapshot
   */

  sendSnapshot(document, action, snapshot) {
    const message = {
      action,
      document,
    }

    if (snapshot) {
      message.snapshot = snapshot
    }

    this.send(message)
  }

  /**
   * Send an `operation` message to the client for a `document` and `action`.
   *
   * @param {String} document
   * @param {String} action
   * @param {Object} operation
   */

  sendOperation(document, action, operation) {
    const message = {
      action,
      document,
      id: operation.id,
      version: operation.version,
    }

    if (operation.data !== undefined) {
      message.data = operation.data
    }

    this.send(message)
  }

  /**
   * Send `operations` to the server for a `document`.
   *
   * @param {String} document
   * @param {String} action
   * @param {Array} operations
   */

  sendOperations(document, action, operations) {
    for (const op of operations) {
      this.sendOperation(document, action, op)
    }
  }

  /**
   * Send an unsubscribe message to the client for a `document`.
   *
   * @param {String} document
   */

  sendUnsubscribe() {
    this.send({
      action: 'unsubscribe',
      document,
    })
  }

  /**
   * On message.
   *
   * @param {Object} message
   * @return {Mixed}
   */

  async onMessage(message) {
    debug('onMessage', this.id, { message })

    checkMessage(message)

    try {
      switch (message.action) {
        case 'connect': return await this.onMessageConnect(message)
        case 'fetch': return await this.onMessageFetch(message)
        case 'submit': return await this.onMessageSubmit(message)
        case 'subscribe': return await this.onMessageSubscribe(message)
        case 'unsubscribe': return await this.onMessageUnsubscribe(message)
      }
    } catch (error) {
      this.send({
        ...message,
        error: {
          code: error.code,
          message: error.message,
          data: error.data,
        }
      })
    }
  }

  /**
   * On "connect" message.
   *
   * @param {Object} message
   * @return {Object}
   */

  async onMessageConnect(message) {
    this.sendConnect()
  }

  /**
   * On "fetch" message.
   *
   * @param {Object} message
   * @return {Object|Void}
   */

  async onMessageFetch(message, options = {}) {
    const { action = 'fetch' } = options
    const { context, server } = this
    const { document, version } = message
    let ops = []
    let snapshot

    if (version == null) {
      snapshot = await server.findSnapshot(document, { context })
    } else {
      ops = await server.listOperations(document, version, undefined, { context })
    }

    this.sendOperations(document, 'broadcast', ops)
    this.sendSnapshot(document, action, snapshot)
  }

  /**
   * On "submit" message.
   *
   * @param {Object} message
   * @return {Object}
   */

  async onMessageSubmit(message) {
    const { context, server } = this
    const { document } = message
    let ops

    try {
      ops = await server.submit(document, message, { context })
    } catch (e) {
      // When a client retries an inflight operation, there's a chance it
      // is a duplicate, but this is expected, and not a "true error".
      if (e.code !== 'operation_duplicate') throw e
    }

    if (ops) {
      this.sendOperations(document, 'broadcast', ops)
    }

    const ack = omit(message, 'data')
    this.sendOperation(document, 'submit', ack)
  }

  /**
   * On "subscribe" message.
   *
   * @param {Object} message
   * @return {Object|Void}
   */

  async onMessageSubscribe(message) {
    const { context, server, subscriptions, id } = this
    const { document, version } = message
    const subscription = await server.subscribe(document, version, { context })
    const previous = subscriptions[document]

    if (previous) {
      previous.push(null)
    }

    subscriptions[document] = subscription
    subscription.on('readable', async () => {
      const op = subscription.read()

      if (op === null) {
        delete subscriptions[document]
        subscription.emit('close')
        return
      }

      // If the operation is from this client, don't send it down to the client
      // because they've already submitted them.
      if (op.id.startsWith(`${id}:`)) return

      // Otherwise, broadcast the op to the client.
      this.sendOperation(document, 'broadcast', op)
    })

    // Call into the fetch handler to fetch the current snapshot or operations.
    await this.onMessageFetch(message, { action: 'subscribe' })
  }

  /**
   * On "unsubscribe" message.
   *
   * @param {Object} message
   */

  async onMessageUnsubscribe(message) {
    const { document, subscriptions } = message
    const sub = subscriptions[document]

    if (sub) {
      sub.push(null)
    }

    this.sendUnsubscribe(document)
  }

}

/**
 * Export.
 *
 * @type {Connection}
 */

export default Connection
