
import Debug from 'debug'
import Emitter from 'events'
import clone from 'lodash/clone'
import isEmpty from 'is-empty'
import { Readable } from 'stream'

import Client from '../client/client'
import Connection from './connection'
import NexusError from '../utils/nexus-error'
import Pubsub from '../adapters/pubsub'
import Store from '../adapters/store'
import Transport from '../adapters/transport'

import {
  checkOperation,
  checkPubsub,
  checkSnapshot,
  checkStore,
  checkTransport,
  checkType,
  checkTypes,
} from '../utils/checkers'

/**
 * Debug.
 *
 * @type {Function}
 */

const debug = Debug('nexus:server')

/**
 * An auto-incrementing ID for subscriptions.
 *
 * @type {Number}
 */

let subscriptionUid = 0

/**
 * Nexus server.
 *
 * @type {Server}
 */

class Server extends Emitter {

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  constructor(config = {}) {
    super()

    const {
      store = new Store(),
      pubsub = new Pubsub(),
      types = [],
      maxRetries = 100,
    } = config

    checkPubsub(pubsub)
    checkStore(store)
    checkTypes(types)

    this.store = store
    this.types = types
    this.pubsub = pubsub
    this.maxRetries = maxRetries
    this.connections = {}
    this.subscriptions = {}
    this.subscribed = {}

    pubsub.on('publish', (id, operation) => {
      if (typeof id !== 'string') {
        throw new NexusError('pubsub_id_invalid', { value: id })
      }

      if (typeof operation !== 'object') {
        throw new NexusError('pubsub_operation_invalid', { value: operation })
      }

      this.broadcast(id, operation)
    })
  }

  /**
   * Initialize a new connection for a `transport` stream with `context`.
   *
   * @param {Stream} transport
   * @param {Object} context
   * @return {Connection}
   */

  async listen(transport, context = {}) {
    debug('listen')

    checkTransport(transport)

    const server = this
    const { store, pubsub, connections } = this

    await Promise.all([
      store.open && store.open(),
      pubsub.open && pubsub.open(),
    ])

    const connection = new Connection({ server, transport, context })
    const { id } = connection

    connections[id] = connection
    connection.once('close', () => {
      delete connections[id]
    })

    return connection
  }

  /**
   * Close the server gracefully.
   */

  async close() {
    debug('close')

    const { store, pubsub, connections } = this

    for (const id in connections) {
      const connection = connections[id]
      connection.close()
    }

    await Promise.all([
      store.close && store.close(),
      pubsub.close && pubsub.close(),
    ])
  }

  /**
   * Create a connected, in-memory client for the server with `context`.
   *
   * @param {Object} context
   * @return {Client}
   */

  async connect(context) {
    const { types } = this
    const transport = new Transport()

    await this.listen(transport.server, context)

    const client = new Client({ types, transport: transport.client })
    return client
  }

  /**
   * Submit an `operation` to a document by `id`.
   *
   * @param {String} id
   * @param {Object} operation
   * @param {Object} options
   * @return {Array}
   */

  async submit(id, operation, options = {}) {
    debug('submit', { id, operation, options })

    checkOperation(operation)

    const { maxRetries } = this
    const { retry = 0, context } = options
    const snapshot = await this.findSnapshot(id, { context })

    checkType(snapshot.type, this.types)

    const type = this.types.find(t => t.name === snapshot.type)
    const operations = []

    // The operation version is always from a previous snapshot version, so it
    // should never be larger than the snapshot.
    if (operation.version > snapshot.version) {
      throw new NexusError('operation_version_mismatch', { operation })
    }

    // If the operation version is older than the snapshot, transform it up to
    // the latest version using the previously stored operations.
    if (operation.version < snapshot.version) {
      const existing = await this.listOperations(id, operation.version, snapshot.version, { context })

      for (const ex of existing) {
        if (operation.id === ex.id) {
          throw new NexusError('operation_duplicate', { operation })
        }

        if (operation.version !== ex.version) {
          throw new NexusError('transform_version_mismatch', { operation, snapshot })
        }

        debug('transform', { id, operation, existing: ex })

        // Transform the operation by the existing one, and bump its version.
        operation.data = type.transform(operation.data, ex.data, 'left')
        operation.version++
        operations.push(ex)
      }
    }

    if (operation.version !== snapshot.version) {
      throw new NexusError('apply_version_mismatch', { operation, snapshot })
    }

    debug('apply', { id, operation, snapshot })

    // Apply the operation to the snapshot, and bump the snapshot's version.
    snapshot.data = type.apply(snapshot.data, operation.data)
    snapshot.version++

    // Attempt to commit the resolve operation and snapshot.
    const success = await this.commit(id, operation, snapshot, { context })

    // Sometimes the commit will fail because another client committed while we
    // were resolving the operation and snapshot. This is normal, so retry.
    if (!success) {
      if (retry <= maxRetries) {
        debug('retry', { id, operation, options })
        return await this.submit(id, operation, { ...options, retry: retry + 1 })
      }

      throw new NexusError('operation_submit_failed', { operation })
    }

    this.publish(id, operation, { context })
    return operations
  }

  /**
   * Commit an `operation` and `snapshot` to the store for a document by `id`,
   * returning a boolean of whether the commit attempt was successful.
   *
   * @param {String} id
   * @param {Object} operation
   * @param {Object} snapshot
   * @param {Object} options
   * @return {Boolean}
   */

  async commit(id, operation, snapshot, options = {}) {
    debug('commit', { id, operation, snapshot, options })

    checkOperation(operation)
    checkSnapshot(snapshot)

    const { context } = options
    const success = await this.store.commit(id, operation, snapshot, { context })

    // Verify the result is valid, because we're going to retry based on it, and
    // we don't want to trigger retry loops when the return value is wrong.
    if (typeof success !== 'boolean') {
      throw new NexusError('store_commit_invalid')
    }

    return success
  }

  /**
   * Find the latest snapshot of a document by `id`.
   *
   * @param {String} id
   * @return {Object}
   */

  async findSnapshot(id, options = {}) {
    debug('findSnapshot', { id, options })

    const { context } = options
    const snapshot = await this.store.findSnapshot(id, { context })

    checkSnapshot(snapshot)
    return snapshot
  }

  /**
   * List all of the operations for a document by `id`, starting `from` version
   * and going up `to` version (non-inclusive).
   *
   * @param {String} id
   * @param {Number} from
   * @param {Number} to
   * @param {Object} options
   * @return {Array}
   */

  async listOperations(id, from, to = Infinity, options = {}) {
    debug('listOperations', { id, from, to, options })

    const { context } = options
    const ops = await this.store.listOperations(id, from, to, { context })
    ops.forEach(op => checkOperation(op))
    return ops
  }

  /**
   * Subscribe to a document by `id` at `version`, broadcasting all future
   * operations published to the document to the subscription stream.
   *
   * @param {String} id
   * @param {Number} version
   * @param {Object} options
   * @return {Readable}
   */

  async subscribe(id, version, options = {}) {
    debug('subscribe', { id, version })

    const { context } = options
    const subs = this.subscriptions[id] || (this.subscriptions[id] = {})
    const subscription = new Readable({ objectMode: true })
    const uid = subscriptionUid++

    subs[uid] = subscription
    subscription._read = () => {}
    subscription.once('close', async () => {
      const s = this.subscriptions[id]
      if (!s) return
      delete s[uid]
      if (isEmpty(s)) await this.unsubscribe(id, { context })
    })

    if (!this.subscribed[id]) {
      await this.pubsub.subscribe(id, { context })
      this.subscribed[id] = true
    }

    return subscription
  }

  /**
   * Unsubscribe from a document by `id`.
   *
   * @param {String} id
   * @param {Object} options
   */

  async unsubscribe(id, options = {}) {
    debug('unsubscribe', { id })

    const { context } = options
    delete this.subscriptions[id]
    delete this.subscribed[id]
    await this.pubsub.unsubscribe(id, { context })
  }

  /**
   * Publish an `operation` to a document by `id`.
   *
   * @param {String} id
   * @param {Object} operation
   * @param {Object} options
   */

  async publish(id, operation, options = {}) {
    debug('publish', { id, operation })

    const { context } = options
    await this.pubsub.publish(id, operation, { context })
  }

  /**
   * Broadcast an `operation` to all subscriptions on a document by `id`.
   *
   * @param {String} id
   * @param {Object} operation
   */

  async broadcast(id, operation) {
    debug('broadcast', { id, operation })

    const subs = this.subscriptions[id]
    if (!subs) return

    for (const key in subs) {
      const sub = subs[key]
      const copy = clone(operation)
      sub.push(copy)
    }
  }

}

/**
 * Export.
 *
 * @type {Server}
 */

export default Server
