
import isEmpty from 'is-empty'

import ACTIONS from '../constants/actions'
import NexusError from '../utils/nexus-error'

/**
 * Check a `message` object.
 *
 * @param {Object} message
 */

function checkMessage(message) {
  if (typeof message !== 'object') {
    throw new NexusError('message_invalid', { value: message })
  }

  if (typeof message.action !== 'string' || !ACTIONS.includes(message.action)) {
    throw new NexusError('message_property_invalid', { message, property: 'action' })
  }
}

/**
 * Check an `operation` object.
 *
 * @param {Object} operation
 */

function checkOperation(operation) {
  if (typeof operation !== 'object') {
    throw new NexusError('operation_invalid', { operation })
  }

  if (typeof operation.id !== 'string') {
    throw new NexusError('operation_property_invalid', { operation, property: 'id' })
  }

  if (typeof operation.version !== 'number') {
    throw new NexusError('operation_property_invalid', { operation, property: 'version' })
  }

  if (typeof operation.data !== 'object') {
    throw new NexusError('operation_property_invalid', { operation, property: 'data' })
  }
}

/**
 * Check a `pubsub` adapter.
 *
 * @param {Object} pubsub
 */

function checkPubsub(pubsub) {
  if (pubsub == null || typeof pubsub !== 'object') {
    throw new Error('The `pubsub` object you passed to the Nexus server is invalid.')
  }

  if (typeof pubsub.on !== 'function') {
    throw new Error('Nexus pubsub adapters must implement the `EventEmitter` interface.')
  }

  if (typeof pubsub.subscribe !== 'function') {
    throw new Error('Nexus pubsub adapters must implement the `subscribe()` method.')
  }

  if (typeof pubsub.unsubscribe !== 'function') {
    throw new Error('Nexus pubsub adapters must implement the `unsubscribe()` method.')
  }

  if (typeof pubsub.publish !== 'function') {
    throw new Error('Nexus pubsub adapters must implement the `publish()` method.')
  }
}

/**
 * Check an `snapshot` object.
 *
 * @param {Object} snapshot
 */

function checkSnapshot(snapshot) {
  if (typeof snapshot !== 'object') {
    throw new NexusError('snapshot_invalid', { snapshot })
  }

  if (typeof snapshot.version !== 'number') {
    throw new NexusError('snapshot_version_invalid', { snapshot, property: 'version' })
  }

  if (typeof snapshot.type !== 'string') {
    throw new NexusError('snapshot_type_invalid', { snapshot, property: 'type' })
  }
}

/**
 * Check a `store` adapter.
 *
 * @param {Object} store
 */

function checkStore(store) {
  if (typeof store !== 'object') {
    throw new Error('The `store` object you passed to the Nexus server is invalid.')
  }

  if (typeof store.commit !== 'function') {
    throw new Error('Nexus store adapters must implement the `commit()` method.')
  }

  if (typeof store.findSnapshot !== 'function') {
    throw new Error('Nexus store adapters must implement the `findSnapshot()` method.')
  }

  if (typeof store.listOperations !== 'function') {
    throw new Error('Nexus store adapters must implement the `listOperations()` method.')
  }
}

/**
 * Check a `transport` adapter.
 *
 * @param {Object} transport
 */

function checkTransport(transport) {
  if (typeof transport !== 'object') {
    throw new Error('The `transport` object you passed to the Nexus server is invalid.')
  }

  if (typeof transport.write !== 'function') {
    throw new Error('Nexus transport adapters must implement the `write()` method.')
  }

  if (typeof transport.read !== 'function') {
    throw new Error('Nexus transport adapters must implement the `read()` method.')
  }

  if (typeof transport.push !== 'function') {
    throw new Error('Nexus transport adapters must implement the `push()` method.')
  }
}

/**
 * Check a `type` string with `types` object.
 *
 * @param {String} type
 * @param {Array} types
 */

function checkType(type, types) {
  if (typeof type !== 'string') {
    throw new NexusError('type_invalid', { type })
  }

  if (!Array.isArray(types) || !types.find(t => t.name === type)) {
    throw new NexusError('type_unknown', { type })
  }
}

/**
 * Check a `types` object.
 *
 * @param {Object} types
 */

function checkTypes(types) {
  if (!Array.isArray(types) || isEmpty(types)) {
    throw new Error('The OT `types` array you passed to Nexus is invalid. It should be an array of OT types.')
  }
}

/**
 * Export.
 *
 * @type {Function}
 */

export {
  checkMessage,
  checkOperation,
  checkPubsub,
  checkStore,
  checkSnapshot,
  checkTransport,
  checkType,
  checkTypes,
}
