
import JSON0 from 'ot-json0' // eslint-disable-line import/no-extraneous-dependencies
import assert from 'assert'
import assertThrows from 'assert-throws-async' // eslint-disable-line import/no-extraneous-dependencies
import { spy } from 'sinon' // eslint-disable-line import/no-extraneous-dependencies

import Connection from '../../lib/server/connection'
import Nexus, { Server, MemoryTransport } from '../..'

/**
 * Types.
 */

const types = [
  JSON0.type
]

/**
 * Tests.
 */

describe('server', () => {
  describe('constructor()', () => {
    it('is a constructor', async () => {
      const nexus = new Nexus({ types })
      assert(nexus instanceof Nexus)
    })

    it('is exposed as `Server`', async () => {
      assert.equal(Server, Nexus)
    })

    it('throws with invalid `pubsub`', async () => {
      assertThrows(
        () => new Nexus({ types, pubsub: {}}),
        Error,
      )
    })

    it('throws with invalid `store`', async () => {
      assertThrows(
        () => new Nexus({ types, store: {}}),
        Error,
      )
    })

    it('throws with invalid `types`', async () => {
      assertThrows(
        () => new Nexus({ types: {}}),
        Error,
      )
    })
  })

  describe('listen()', () => {
    it('is a method', async () => {
      const nexus = new Nexus({ types })
      assert.equal(typeof nexus.listen, 'function')
    })

    it('calls `open()` on pubsub', async () => {
      const pubsub = pubsubSpy()
      const nexus = new Nexus({ types, pubsub })
      const transport = new MemoryTransport()
      await nexus.listen(transport.server)
      assert(pubsub.open.called)
    })

    it('calls `open()` on store', async () => {
      const store = storeSpy()
      const nexus = new Nexus({ types, store })
      const transport = new MemoryTransport()
      await nexus.listen(transport.server)
      assert(store.open.called)
    })

    it('returns a `Connection` object', async () => {
      const nexus = new Nexus({ types })
      const transport = new MemoryTransport()
      const connection = await nexus.listen(transport.server)
      assert(connection instanceof Connection)
    })

    it('attaches `context` to the connection', async () => {
      const nexus = new Nexus({ types })
      const transport = new MemoryTransport()
      const context = {}
      const connection = await nexus.listen(transport.server, context)
      assert.equal(connection.context, context)
    })

    it('throws with invalid `transport`', async () => {
      const nexus = new Nexus({ types })
      assertThrows(
        async () => await nexus.listen({}),
        Error,
      )
    })
  })

  describe('close()', () => {
    it('calls `close()` on pubsub', async () => {
      const pubsub = pubsubSpy()
      const nexus = new Nexus({ types, pubsub })
      await nexus.close()
      assert(pubsub.close.called)
    })

    it('calls `close()` on store', async () => {
      const store = storeSpy()
      const nexus = new Nexus({ types, store })
      await nexus.close()
      assert(store.close.called)
    })
  })
})

function pubsubSpy() {
  return {
    open: spy(),
    close: spy(),
    on: spy(),
    removeListener: spy(),
    subscribe: spy(),
    unsubscribe: spy(),
    publish: spy(),
  }
}

function storeSpy() {
  return {
    open: spy(),
    close: spy(),
    commit: spy(),
    findSnapshot: spy(),
    listOperations: spy(),
  }
}
